#import "StructureBlend.h"
#import "QCGLCIImage.h"

static CIKernel	*frameBlenderKernel = nil;
static CIFilter	*frameBlenderFilter = nil;

@implementation StructureBlend : QCPatch

+ (int)executionMode
{
	return 2;
}
+ (BOOL)allowsSubpatches
{
	return NO;
}
+ (int)timeMode
{
	return 1;
}

- (id)initWithIdentifier:(id)fp8
{
	id z=[super initWithIdentifier:fp8];

	if(z)
	{
		// ordinal/normalized
		[inputOrderMode setMaxIndexValue:1];
		// clamp/repeat
		[inputWrapMode setMaxIndexValue:1];
		// nearest neighbor/linear
		[inputBlendMode setMaxIndexValue:1];
		if(frameBlenderFilter == nil)
		{
			NSString *kernelSource = 
				@"kernel vec4 frameBlender(sampler a, sampler b,float alpha){return sample(a,samplerCoord(a))*(1.0-alpha)+sample(b,samplerCoord(b))*alpha;}";
			//	@"kernel vec4 frameBlender(sampler a){return sample(a, samplerCoord(a));}";
			//	@"kernel vec4 frameBlender(sampler a, float alpha){return samplerCoord(a, samplerCoord(a))*alpha;}";
			NSArray *kernels = [CIKernel kernelsWithString:kernelSource];
			frameBlenderKernel = [kernels objectAtIndex:0];
			[frameBlenderKernel retain];
			//NSLog(@"frameBlender: (%@) %@",[frameBlender class], frameBlender);
			
			frameBlenderFilter = [[CIFilter alloc] init];
			//NSLog(@"fbf is (%@) %@",[frameBlenderFilter class], frameBlenderFilter);
		}
	}

	return z;
}


- (id)setup:(QCOpenGLContext *)context
{


	return context;
}
- (void)cleanup:(QCOpenGLContext *)context
{


}


- (void)enable:(QCOpenGLContext *)context
{


}
- (void)disable:(QCOpenGLContext *)context
{


}

id blendObjects(id a, id b, float alpha)
{
	id newObject = nil;
	// can only blend identical objects
	if([a class] != [b class])
		return a;
	if(alpha <= 0.0)
		return a;
	else if(alpha >= 1.0)
		return b;
	
	/* Special considerations: if you're allocating something new, mark it as autorelease. */
	if([a isKindOfClass:[NSNumber class]] == TRUE)
	{
		newObject = [[NSNumber alloc] initWithDouble:([a floatValue]*(1.0-alpha))+([b floatValue]*alpha)];
		[newObject autorelease];
	}
	else if([a isKindOfClass:[NSColor class]])
	{
		//NSLog(@"Blending colors!");
		//newObject = [b copy];
		//[newObject blendedColorWithFraction:alpha ofColor:a];
		// I think QC only uses RGB colorspace.  if not, this will throw an exception.
		float r1,g1,b1,a1;
		float r2,g2,b2,a2;
		[a getRed:&r1 green:&g1 blue:&b1 alpha:&a1];
		[b getRed:&r2 green:&g2 blue:&b2 alpha:&a2];
		newObject = [NSColor colorWithCalibratedRed: r1*(1.0-alpha)+r2*alpha
				green: g1*(1.0-alpha)+g2*alpha
				blue:  b1*(1.0-alpha)+b2*alpha
				alpha: a1*(1.0-alpha)+a2*alpha];
	}
	else if([a isKindOfClass:[QCGLImage class]])
	{
		CIVector *vec = [CIVector vectorWithX:alpha];
		// HW-accelerated frame blender.  Nice
		//NSLog(@"a: %x  b:%x",a,b);
		//NSLog(@"blender kernel is (%@) %@",[frameBlenderKernel class],frameBlenderKernel);
		//NSLog(@"blender filter is (%@) %@",[frameBlenderFilter class],frameBlenderFilter);
		newObject = [frameBlenderFilter apply:frameBlenderKernel, [a CIImage], [b CIImage], vec, @"definition", [[a CIImage] definition], nil];
		//newObject = [frameBlenderFilter apply:frameBlenderKernel, [a CIImage],[b CIImage], alpha, nil];
		//NSLog(@"applying filter returned %@",newObject);
	}
	else if([a isKindOfClass:[NSString class]])
	{
		// anyone know any slick text blending algorithms? ;)
		return a;
	}
	else
	{
		// handy for extending the blender to new objects
		//NSLog(@"don't know how to blend (%@) %@",[a class],a);
	}
	
	
	if(newObject != nil)
		return newObject;
	
	/* if it's an unsupported type, just return nearest neighbor */
	/* note: this isn't real nearest neighbor yet (0.9 will return a, not b)*/
	return a;
}

- (BOOL)execute:(QCOpenGLContext *)context time:(double)time arguments:(NSDictionary *)arguments
{
	NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
	
	QCStructure *structure = [inputStruct structureValue];
	
	if(structure == nil || [structure count] == 0)	// if there isn't a buffer, or it's empty; there's no point evaluating more.
	{
		[outputVirtual setRawValue:nil];
		return YES;
	}
	
	// Note: index is a floating point value.  Never check index with equality only, it must have greater-than or less-than to be useful.
	double index = [inputIndex doubleValue];
	double frac;
	
	// our fractional part is different for normalized operation.
	if([inputOrderMode indexValue] == 0)
		frac  = index - floor(index);
	else
		frac = fmod(index, 1.0/([structure count]-1)) * ([structure count]-1.0);
	
	switch([inputOrderMode indexValue])
	{
		default:
		case 0:	// ordinal
			switch([inputWrapMode indexValue])
			{
				default:
				case 0:	// clamp
					if(index <= 0.0)
						[outputVirtual setRawValue:[structure memberAtIndex:0]];
					else if(index >= [structure count]-1)
						[outputVirtual setRawValue:[structure memberAtIndex:[structure count]-1]];
					else
						if([inputBlendMode indexValue])
							[outputVirtual setRawValue:
								blendObjects([structure memberAtIndex:(int)index],[structure memberAtIndex:((int)index)+1],frac)];
						else
							[outputVirtual setRawValue:[structure memberAtIndex:(int)index]];
					break;
				case 1:	// repeat
					// special case: if there's only 1 element in the array, our fmod will try to 
					// divide by zero.  repeat is trivial in count==1 case anyway :)
					if([structure count] == 1)
						[outputVirtual setRawValue:[structure memberAtIndex:0]];
					else
					{
						if( index > [structure count]-1.0 || index < 0.0 )
							index = fmod(index,[structure count] - 1.0);
						if(index < 0.0)	// -3%5 is -3, not 2. -7%5 is -2, not 3 
							index += (double)[structure count];
							
						if([inputBlendMode indexValue])
							[outputVirtual setRawValue:
								blendObjects([structure memberAtIndex:(int)index],[structure memberAtIndex:((int)index)+1],frac)];
						else
							[outputVirtual setRawValue:[structure memberAtIndex:(int)index]];
					}
					break;
			}
			break;
		case 1:	// normalized
			switch([inputWrapMode indexValue])
			{
				default:
				case 0:	// clamp
					if(index <= 0.0)
						[outputVirtual setRawValue:[structure memberAtIndex:0]];
					else if(index >= 1.0)
						[outputVirtual setRawValue:[structure memberAtIndex:[structure count]-1]];
					else
						if([inputBlendMode indexValue])
							[outputVirtual setRawValue:
								blendObjects(
									[structure memberAtIndex:(int) (index * ((double)[structure count] - 1.0))],
									[structure memberAtIndex:(int) (index * ((double)[structure count] - 1.0))+1],
									frac)];
						else
							[outputVirtual setRawValue:[structure memberAtIndex:(int) (index * ((double)[structure count] - 1.0))]];
					break;
				case 1:	// repeat
					// special case: if there's only 1 element in the array, our fmod will try to 
					// divide by zero.  repeat is trivial in count==1 case anyway :)
					if([structure count] == 1)
						[outputVirtual setRawValue:[structure memberAtIndex:0]];
					else
					{
						if(index > 1.0 || index < 0.0)
							index = fmod(index, 1.0);
						if(index < 0.0)	// -3%5 is -3, not 2. -7%5 is -2, not 3 
							index += 1.0;
						if([inputBlendMode indexValue])
							[outputVirtual setRawValue:
								blendObjects(
									[structure memberAtIndex:(int) (index * ((double)[structure count] - 1.0))],
									[structure memberAtIndex:(int) (index * ((double)[structure count] - 1.0))+1],
									frac)];
						else
							[outputVirtual setRawValue: [structure memberAtIndex:(int) (index * ((double)[structure count] - 1.0))]];

						//[outputVirtual setRawValue:[structure memberAtIndex:
						//				(int) (index * ((double)[structure count] - 1.0)) 
						//]];
					}
					break;
			}
			break;
	}

	[pool release];
	
	return YES;
}

@end
